import json
import os
import urllib.request
from collections import defaultdict
from typing import List, Tuple, Optional, Iterator, Dict, Any

import semver

from skjold.models import SecurityAdvisory, SecurityAdvisorySource
from skjold.tasks import register_source


class GithubSecurityAdvisory(SecurityAdvisory):
    _json: Dict

    @classmethod
    def using(cls, json_: dict) -> "GithubSecurityAdvisory":
        obj = cls()
        obj._json = json_
        return obj

    @property
    def identifier(self) -> str:
        return str(self.__advisory["ghsaId"])

    @property
    def source(self) -> str:
        return "github"

    @property
    def references(self) -> List[str]:
        return [reference["url"] for reference in self.__advisory["references"]]

    @property
    def package_name(self) -> str:
        return str(self._json["node"]["package"]["name"])

    @property
    def ecosystem(self) -> str:
        return str(self._json["node"]["package"]["ecosystem"])

    @property
    def severity(self) -> str:
        return str(self._json["node"]["severity"])

    @property
    def __advisory(self) -> Any:
        return self._json["node"]["advisory"]

    @property
    def summary(self) -> str:
        return str(self.__advisory["summary"])

    @property
    def first_patched_version(self) -> str:
        return str(self._json["node"]["firstPatchedVersion"]["identifier"])

    @property
    def vulnerable_version_range(self) -> semver.VersionConstraint:
        return semver.parse_constraint(self._json["node"]["vulnerableVersionRange"])

    @property
    def vulnerable_versions(self) -> str:
        return str(self.vulnerable_version_range)

    def is_affected(self, version: str) -> bool:
        version = semver.Version.parse(version)
        if self.vulnerable_version_range.allows(version):
            return True
        return False

    @property
    def url(self) -> str:
        return f"https://github.com/advisories/{self.identifier}"


def _query_github_graphql(
    first: int = 10, after: Optional[str] = None
) -> Tuple[int, str, bool, dict]:
    _after = after and f'"{after}"' or "null"
    _limit = first and int(first) or 1
    _api_token = os.environ["SKJOLD_GITHUB_API_TOKEN"]

    query = f"""
    {{
        securityVulnerabilities(first: {_limit}, after: {_after}, ecosystem: PIP, orderBy: {{ field: UPDATED_AT, direction: DESC }}) {{
            pageInfo {{
                startCursor
                hasNextPage
                endCursor
            }}
            totalCount
            edges {{
                node {{
                    advisory {{
                        ghsaId
                        publishedAt
                        references {{
                            url
                        }}
                        summary
                    }}
                    firstPatchedVersion {{
                        identifier
                    }}
                    package {{
                        ecosystem
                        name
                    }}
                    severity
                    updatedAt
                    vulnerableVersionRange
                }}
            }}
        }}
    }}
    """
    payload = json.dumps({"query": query}).encode("utf-8")
    request_ = urllib.request.Request(
        url="https://api.github.com/graphql",
        data=payload,
        headers={
            "Accept": "application/json",
            "Authorization": f"Bearer {_api_token}",
            "Content-Type": "application/json; charset=utf-8",
        },
    )
    with urllib.request.urlopen(request_) as response:
        _data = json.loads(response.read())

    has_next = _data["data"]["securityVulnerabilities"]["pageInfo"]["hasNextPage"]
    cursor = _data["data"]["securityVulnerabilities"]["pageInfo"]["endCursor"]
    data = _data["data"]["securityVulnerabilities"]["edges"]
    total_count = int(_data["data"]["securityVulnerabilities"]["totalCount"])

    return total_count, cursor, has_next, data


def _fetch_github_security_advisories(
    limit: int = 100,
) -> Iterator[Tuple[int, str, bool, dict]]:
    total_count, cursor, has_next, data = _query_github_graphql(limit, None)
    yield from data

    while has_next:
        total_count, cursor, has_next, data = _query_github_graphql(limit, cursor)
        yield from data


class Github(SecurityAdvisorySource):
    _name = "github"

    @property
    def name(self) -> str:
        return self._name

    @property
    def total_count(self) -> int:
        return len(self._advisories.keys())

    def populate_from_cache(self) -> None:
        self._advisories = defaultdict(list)

        with open(self.path, "rb") as fh:
            for item in json.load(fh):
                obj = GithubSecurityAdvisory.using(item)
                self._advisories[obj.package_name].append(obj)

    @property
    def path(self) -> str:
        return os.path.join(self._cache_dir, "github.cache")

    def update(self) -> None:
        with open(self.path, "w") as fh:
            json.dump(list(_fetch_github_security_advisories()), fh)

    def has_security_advisory_for(self, package_name: str) -> bool:
        return package_name.strip() in self.advisories.keys()

    def is_vulnerable_package(
        self, package_name: str, package_version: str
    ) -> Tuple[bool, List[SecurityAdvisory]]:
        advisories = []
        for candidate in self.advisories[package_name]:
            if candidate.is_affected(package_version):
                advisories.append(candidate)

        return len(advisories) > 0, advisories


register_source("github", Github)



import json
import os
from typing import TextIO, MutableMapping, Callable, Optional, Iterator

import click
import toml

from skjold.models import PackageList, Package, SkjoldException
from skjold.tasks import Configuration


def read_poetry_lock_from(file: TextIO) -> Iterator[Package]:
    """Reads a poetry.lock given by path and yields 'package==version' items."""
    doc = toml.loads(file.read())
    for package in doc.get("package", []):
        yield package["name"], package["version"]


def read_pipfile_lock_from(file: TextIO) -> Iterator[Package]:
    """Reads a Pipfile.lock given by path and yields 'package==version' items."""
    json_ = json.load(file)
    for namespace in ["develop", "default"]:
        for package_name in json_[namespace].keys():
            package_info = json_[namespace][package_name]

            if "version" not in package_info:
                continue

            pinned_package_version = package_info["version"]
            if not pinned_package_version.startswith("=="):
                raise SkjoldException(
                    f"Unexpected value for pinned version '{pinned_package_version}'!"
                )

            package_version = pinned_package_version.replace("==", "")
            yield package_name, package_version


def read_requirements_txt_from(file: TextIO) -> Iterator[Package]:
    """Reads a requirements.txt given by path and yields 'package==version' items."""
    for line in file.readlines():
        # Skip empty lines or lines only containing a hash.
        if line.strip().startswith("--hash") or not len(line.strip()):
            continue
        # Skip lines only containing editable packages.
        if line.strip().startswith("-e"):
            continue
        # Skip comment lines.
        if line.strip().startswith("#"):
            continue

        try:
            line = line.split(";")[0]
            package_name, package_version = line.strip().split(" ")[0].split("==")
            yield package_name, package_version
        except (ValueError, KeyError):
            click.secho("Warning! ", err=True, nl=False, fg="yellow")
            click.secho(
                f"Unable extract package and version from '{line.strip()}'. Skipping!",
                err=True,
            )
            continue


class Format:  # pragma: no cover
    POETRY: str = "poetry.lock"
    REQUIREMENTS: str = "requirements.txt"
    PIPENV: str = "Pipfile.lock"

    SUPPORTED_FORMATS: MutableMapping[str, Callable] = {
        POETRY: read_poetry_lock_from,
        REQUIREMENTS: read_requirements_txt_from,
        PIPENV: read_pipfile_lock_from,
    }


def extract_package_list_from(
    configuration: Configuration, file: TextIO, format_: Optional[str] = None
) -> PackageList:
    """Extracts the list of tuples containing package name and version."""
    filename = os.path.basename(file.name)

    if not format_ or filename in Format.SUPPORTED_FORMATS.keys():
        format_ = filename
        if configuration.verbose:
            click.secho(f"Assuming '{format_}' from filename.", err=True)

    reader_func = Format.SUPPORTED_FORMATS.get(format_, None)
    if not reader_func:
        raise SkjoldException(f"Unsupported file or format '{format_}'!")

    _packages = []
    for package in reader_func(file):
        _packages.append(package)

    return _packages



import abc
import os
import time
from abc import ABCMeta, abstractmethod
from typing import List, Tuple, MutableMapping, Sequence, Optional


class SkjoldException(Exception):
    pass


Package = Tuple[str, str]
PackageList = Sequence[Package]


class SecurityAdvisory(metaclass=abc.ABCMeta):
    @property
    @abstractmethod
    def identifier(self) -> str:
        """Return this advisories unique identifier."""
        raise NotImplementedError

    @property
    @abstractmethod
    def source(self) -> str:
        """Return the data source the advisory comes from."""
        raise NotImplementedError

    @property
    @abstractmethod
    def package_name(self) -> str:
        """Return package name of the affected package."""
        raise NotImplementedError

    @property
    @abstractmethod
    def url(self) -> str:
        """Return direct link to more information for a given advisory."""
        raise NotImplementedError

    @property
    @abstractmethod
    def references(self) -> List[str]:
        """Return list of references for this advisory."""
        raise NotImplementedError

    @property
    @abstractmethod
    def summary(self) -> str:
        """Return string containing a short summary of the advisory."""
        raise NotImplementedError

    @property
    @abstractmethod
    def severity(self) -> str:
        """Return the severity level of the advisory/underlying vulnerability."""
        raise NotImplementedError

    @property
    @abstractmethod
    def vulnerable_versions(self) -> str:
        """Get string representation of the affected version ranges."""
        raise NotImplementedError

    @abstractmethod
    def is_affected(self, version: str) -> bool:
        """Return True if the given version is within the affected version range. False otherwise."""
        raise NotImplementedError


SecurityAdvisoryList = List[SecurityAdvisory]


def is_outdated(path: str, max_age: int = 3600) -> bool:
    """Return True if the given file's mtime exceeds 'max_age'. False otherwise."""
    last_modified = int(os.path.getmtime(path))
    diff = int(time.time()) - last_modified
    return diff >= max_age


class SecurityAdvisorySource(metaclass=ABCMeta):
    _advisories: MutableMapping[str, SecurityAdvisoryList] = {}
    _cache_dir: str
    _cache_expires: int
    _name: str

    def __init__(self, cache_dir: str, cache_expires: int = 0) -> None:
        self._cache_dir = cache_dir
        self._cache_expires = cache_expires

    @property
    @abstractmethod
    def name(self) -> str:
        """Return name of this source."""
        raise NotImplementedError

    @property
    def advisories(self) -> MutableMapping[str, SecurityAdvisoryList]:
        """Return list of SecurityAdvisories from the given source."""
        if self.requires_update:
            self.update()

        if not len(self._advisories):
            self.populate_from_cache()

        return self._advisories

    @property
    def requires_update(self) -> bool:
        """ Return True if the source should be updated. False otherwise. """
        if self.path is None:
            return True

        if not os.path.exists(self.path):
            return True

        return is_outdated(self.path, self._cache_expires)

    @property
    @abstractmethod
    def path(self) -> Optional[str]:
        """Return path to local database download."""
        raise NotImplementedError

    @property
    @abstractmethod
    def total_count(self) -> int:
        """Return number of total security advisories."""
        raise NotImplementedError

    @abstractmethod
    def update(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def populate_from_cache(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def is_vulnerable_package(
        self, package_name: str, package_version: str
    ) -> Tuple[bool, Sequence[SecurityAdvisory]]:
        raise NotImplementedError

    @abstractmethod
    def has_security_advisory_for(self, package_name: str) -> bool:
        raise NotImplementedError

    def get_security_advisories(
        self,
    ) -> MutableMapping[str, SecurityAdvisoryList]:
        return self.advisories



from .github import Github
from .pyup import PyUp
from .gemnasium import Gemnasium
from .pypa import PyPAAdvisoryDB
from .osv import OSV

__all__ = ("Github", "PyUp", "Gemnasium", "OSV", "PyPAAdvisoryDB")



import math
from typing import Dict, Any, Union


def round_up(n: float, decimals: int = 1) -> float:
    multiplier = 10 ** decimals
    return float(math.ceil(n * multiplier) / multiplier)


class CVSS2:

    _basic: Dict[str, str] = {}
    _fields_basic_group = frozenset({"AV", "AC", "Au", "C", "I", "A"})
    _metrics: Dict[str, Any] = {
        "AV": {"N": 1.0, "A": 0.646, "L": 0.395},
        "AC": {"L": 0.71, "M": 0.61, "H": 0.35},
        "Au": {"M": 0.45, "S": 0.56, "N": 0.704},
        "C": {"N": 0.0, "P": 0.275, "C": 0.660},
        "I": {"N": 0.0, "P": 0.275, "C": 0.660},
        "A": {"N": 0.0, "P": 0.275, "C": 0.660},
    }

    @classmethod
    def using(cls, vector: str) -> "CVSS2":
        kv = map(lambda v: v.split(":"), vector.strip().split("/"))
        basic_group = filter(lambda metric: metric[0] in CVSS2._fields_basic_group, kv)

        obj = CVSS2()
        obj._basic = {v[0]: v[1] for v in basic_group}
        return obj

    @property
    def _impact_subscore(self) -> float:
        confidentiality = self._basic["C"]
        integrity = self._basic["I"]
        availability = self._basic["A"]
        return 1 - (
            (1 - float(self._metrics["C"][confidentiality]))
            * (1 - float(self._metrics["I"][integrity]))
            * (1 - float(self._metrics["A"][availability]))
        )

    @property
    def exploitability_score(self) -> float:
        """Calculate the exploitability score."""
        av = float(self._metrics["AV"][self._basic["AV"]])
        ac = float(self._metrics["AC"][self._basic["AC"]])
        au = float(self._metrics["Au"][self._basic["Au"]])
        return 20.0 * av * ac * au

    @property
    def impact_score(self) -> float:
        """Calculate the impact score."""
        return 10.41 * self._impact_subscore

    @property
    def score(self) -> float:
        """Return the CVSS 2.0 base score."""
        if self.impact_score <= 0:
            return 0.0

        return round(
            1.176
            * ((0.6 * self.impact_score) + (0.4 * self.exploitability_score) - 1.5),
            1,
        )

    @property
    def severity(self) -> str:
        """Return severity level for based on the CVSS 2.0 base score."""
        x = self.score
        if x < 0.001:
            return "NONE"
        elif x < 4:
            return "LOW"
        elif x < 7:
            return "MEDIUM"
        else:
            return "HIGH"


class CVSS3:

    _basic: Dict[str, str] = {}
    _fields_basic_group = frozenset({"AV", "AC", "PR", "UI", "S", "C", "I", "A"})
    _metrics: Dict[str, Any] = {
        "AV": {"N": 0.85, "A": 0.62, "L": 0.55, "P": 0.2},
        "AC": {"L": 0.77, "H": 0.44},
        "PR": {
            "C": {"N": 0.85, "L": 0.68, "H": 0.50},
            "U": {"N": 0.85, "L": 0.62, "H": 0.27},
        },
        "UI": {"N": 0.85, "R": 0.62},
        "C": {"H": 0.56, "L": 0.22, "N": 0.0},
        "I": {"H": 0.56, "L": 0.22, "N": 0.0},
        "A": {"H": 0.56, "L": 0.22, "N": 0.0},
    }

    @classmethod
    def using(cls, vector: str) -> "CVSS3":
        kv = map(lambda v: v.split(":"), vector.strip().upper().split("/"))
        basic_group = filter(lambda metric: metric[0] in cls._fields_basic_group, kv)

        obj = CVSS3()
        obj._basic = {v[0]: v[1] for v in basic_group}
        return obj

    @property
    def scope(self) -> str:
        return self._basic["S"]

    @property
    def _impact_subscore(self) -> float:
        confidentiality = self._basic["C"]
        integrity = self._basic["I"]
        availability = self._basic["A"]
        return 1 - (
            (1 - float(self._metrics["C"][confidentiality]))
            * (1 - float(self._metrics["I"][integrity]))
            * (1 - float(self._metrics["A"][availability]))
        )

    @property
    def exploitability_score(self) -> float:
        """Calculate the exploitability score."""
        av = float(self._metrics["AV"][self._basic["AV"]])
        ac = float(self._metrics["AC"][self._basic["AC"]])
        pr = float(self._metrics["PR"][self._basic["S"]][self._basic["PR"]])
        ui = float(self._metrics["UI"][self._basic["UI"]])

        return 8.22 * av * ac * pr * ui

    @property
    def impact_score(self) -> float:
        """Calculate the impact score."""
        impact_subscore = self._impact_subscore
        if self.scope == "U":
            return 6.42 * impact_subscore

        return 7.52 * (impact_subscore - 0.029) - 3.25 * (impact_subscore - 0.02) ** 15

    @property
    def score(self) -> float:
        """Return the CVSS 3.0 base score."""
        if self.impact_score <= 0:
            return 0.0

        if self.scope == "U":
            return round_up(min((self.impact_score + self.exploitability_score), 10.0))

        return round_up(
            min(1.08 * (self.impact_score + self.exploitability_score), 10.0)
        )

    @property
    def severity(self) -> str:
        """Return severity level for based on the CVSS 3.0 base score."""
        x = self.score
        if x < 0.001:
            return "NONE"
        elif x < 4.0:
            return "LOW"
        elif x < 7.0:
            return "MEDIUM"
        elif x < 9.0:
            return "HIGH"
        else:
            return "CRITICAL"


def parse_cvss(vector: str) -> Union[CVSS2, CVSS3]:
    if vector.startswith("CVSS:3"):
        return CVSS3.using(vector)
    return CVSS2.using(vector)



import datetime
import json
import os
import urllib.request
from collections import defaultdict
from typing import Tuple, List, Dict, Callable, Union, Any

import semver

from skjold.models import SecurityAdvisorySource, SecurityAdvisory
from skjold.tasks import register_source


class PyUpSecurityAdvisory(SecurityAdvisory):
    _json: Dict[str, str]

    @classmethod
    def using(cls, name: str, json_: dict) -> "PyUpSecurityAdvisory":
        obj = cls()
        obj._json = json_
        obj._json["name"] = name
        return obj

    @property
    def identifier(self) -> str:
        return self._json["id"]

    @property
    def source(self) -> str:
        return "pyup"

    @property
    def severity(self) -> str:
        return "UNKNOWN"

    @property
    def url(self) -> str:
        return f"https://pyup.io/{self.identifier}"

    @property
    def references(self) -> List[str]:
        return []

    @property
    def package_name(self) -> str:
        return self._json["name"]

    @property
    def summary(self) -> str:
        return self._json["advisory"]

    @property
    def vulnerable_version_range(self) -> List[semver.VersionConstraint]:
        return [semver.parse_constraint(v) for v in self._json["specs"]]

    @property
    def vulnerable_versions(self) -> str:
        return ",".join([str(x) for x in self.vulnerable_version_range])

    def is_affected(self, version: str) -> bool:
        version_ = semver.Version.parse(version)
        allows_: Callable[[semver.VersionConstraint], bool] = (
            lambda x: True if x.allows(version_) else False
        )
        # affected_versions = map(lambda x: x.allows(version), self.vulnerable_version_range)
        affected_versions = map(allows_, self.vulnerable_version_range)
        return any(affected_versions)


class PyUp(SecurityAdvisorySource):

    _url: str = "https://raw.githubusercontent.com/pyupio/safety-db/master/data/insecure_full.json"
    _name: str = "pyup"
    _metadata: Dict[str, Union[str, int]] = {}

    @property
    def name(self) -> str:
        return self._name

    @property
    def total_count(self) -> int:
        return len(self._advisories.keys())

    @property
    def path(self) -> str:
        return os.path.join(self._cache_dir, "pyup.cache")

    @property
    def last_updated_at(self) -> datetime.datetime:
        timestamp = int(self._metadata["timestamp"])
        return datetime.datetime.utcfromtimestamp(timestamp)

    def _load_cache(self) -> Any:
        with open(self.path, "rb") as fh:
            json_ = json.load(fh)
            return json_

    def populate_from_cache(self) -> None:
        self._advisories = defaultdict(list)

        cache_ = self._load_cache()
        for package_name, advisories in cache_.items():
            if package_name in {"$meta"}:
                self._metadata = advisories
                continue

            for advisory in advisories:
                obj = PyUpSecurityAdvisory.using(package_name, advisory)
                self._advisories[obj.package_name].append(obj)

    def update(self) -> None:
        request_ = urllib.request.Request(
            url=self._url,
            headers={"Accept": "application/json"},
        )
        with urllib.request.urlopen(request_) as response:
            json_ = json.loads(response.read())

        with open(self.path, "w") as fh:
            json.dump(json_, fh)

    def has_security_advisory_for(self, package_name: str) -> bool:
        return package_name.strip() in self.advisories.keys()

    def is_vulnerable_package(
        self, package_name: str, package_version: str
    ) -> Tuple[bool, List[SecurityAdvisory]]:
        advisories = []
        for candidate in self.advisories[package_name]:
            if candidate.is_affected(package_version):
                advisories.append(candidate)

        return len(advisories) > 0, advisories


register_source("pyup", PyUp)



"""Contains actual task implementations that can be either called directly or via the click cli."""
import json
import os
import textwrap
from typing import List, MutableMapping, Type, AbstractSet, Union, Dict, Any, Tuple, Set

import click
import toml

from skjold.models import SecurityAdvisorySource, PackageList, SkjoldException
from skjold.ignore import SkjoldIgnore

_sources: MutableMapping[str, Type[SecurityAdvisorySource]] = {}


def default_from_context(attr: str, cls: object) -> Type[click.Option]:
    class OptionDefaultFromContext(click.Option):
        def get_default(self, ctx: Any, call: bool = False) -> Any:
            self.default = getattr(ctx.find_object(cls), attr)
            return super().get_default(ctx, call)

    return OptionDefaultFromContext


def get_configuration_from_toml(filename: str) -> Any:
    """Return the tool.skjold section from the given pyproject.toml location."""
    document = toml.load(filename)
    section = document.get("tool", {}).get("skjold", {})
    return section


class Configuration:
    sources: List[str] = []  # Advisory sources enabled by default.
    report_only: bool = (
        False  # Return non-zero exit code when vulnerabilities are found.
    )
    report_format: str = "cli"  # Output parsable JSON instead of stupid colors.
    cache_dir: str = ".skjold_cache"  # Cache location.
    cache_expires: int = 12 * 3600  # Cache maximum age.
    ignore_file: str = ".skjoldignore"  # Default ignore file.
    verbose: bool = False  # Be verbose when processing package list.

    def use(self, config: Dict) -> None:
        self.sources = config.get("sources", self.sources)
        self.report_only = config.get("report_only", self.report_only)
        self.report_format = config.get("report_format", self.report_format)
        # Configure cache_dir selection: ENV > pyproject.toml > default(posix).
        self.cache_dir = os.environ.get(
            "SKJOLD_CACHE_DIR", config.get("cache_dir", self.default_cache_dir)
        )
        self.cache_expires = config.get("cache_expires", self.cache_expires)
        self.ignore_file = os.environ.get(
            "SKJOLD_IGNORE_FILE", config.get("ignore_file", self.ignore_file)
        )
        # self.verbose = bool(config.get("verbose", self.verbose))

        # Sources
        if not len(self.sources):
            click.secho(
                "Warning: No advisory sources configured!", err=True, fg="yellow"
            )
        for source_name in self.sources:
            if not is_registered_source(source_name):
                raise click.ClickException(
                    f"Source with name '{source_name}' does not exist!"
                )

    @property
    def app_home(self) -> str:
        return str(click.get_app_dir("skjold", roaming=False, force_posix=True))

    @property
    def default_cache_dir(self) -> str:
        return os.path.join(self.app_home, "cache")

    @property
    def available_sources(self) -> AbstractSet[str]:
        """Return list of available sources by name."""
        return _sources.keys()

    def as_dict(self) -> MutableMapping[str, Union[bool, str, int, List]]:
        """Return dictionary representation of configuration object."""
        return {
            "sources": self.sources,
            "report_only": self.report_only,
            "report_format": self.report_format,
            "verbose": self.verbose,
            "cache_dir": self.cache_dir,
            "cache_expires": self.cache_expires,
            "ignore_file": self.ignore_file,
        }


def register_source(new_source_name: str, source: Type[SecurityAdvisorySource]) -> None:
    """Registers a new source by name. Throws Exception otherwise."""
    if not new_source_name or new_source_name in _sources:
        raise SkjoldException(
            f"A source named '{new_source_name}' appears to be already registered!"
        )

    _sources[new_source_name] = source


def get_registered_sources() -> AbstractSet[str]:
    """Return list of keys for registered advisory sources."""
    return _sources.keys()


def is_registered_source(name: str) -> bool:
    """Return True if a resource by the given name exists. False otherwise."""
    return name in _sources.keys()


def print_configuration(configuration: Configuration, stderr: bool = True) -> None:
    """Prints the currently active configuration for skjold to stdout and exits."""
    for key, value in configuration.as_dict().items():
        click.secho(key, fg="white", nl=False, err=stderr)
        click.secho(": ", nl=False, err=stderr)
        click.secho(str(value), fg="white", nl=False, err=stderr)
        click.secho("", err=stderr)


def report(
    configuration: Configuration, findings: List[Dict[str, Any]]
) -> Tuple[Set[str], List[str]]:
    """Renders (ignored) findings and list of vulnerable packages to stdout and prints a short summary to stderr."""
    vulnerable_packages, ignored_findings = set({}), []
    for finding in findings:
        if finding["ignored"]["ignored"]:
            ignored_findings.append(finding["identifier"])
        else:
            vulnerable_packages.add(finding["name"])

    if configuration.report_format == "json":
        click.echo(json.dumps(findings, indent=2))
        return vulnerable_packages, ignored_findings

    for finding in findings:
        # https://nvd.nist.gov/vuln-metrics/cvss
        _color = {
            "NONE": "white",
            "LOW": "yellow",
            "MODERATE": "yellow",  # Github
            "MEDIUM": "yellow",  # CVSS
            "HIGH": "red",
            "CRITICAL": "red",
            "UNKNOWN": "red",
        }.get(finding["severity"])

        if finding["ignored"]["ignored"]:
            click.secho("")
            click.secho(finding["name"], fg="white", nl=False)
            click.secho("==", nl=False)
            click.secho(finding["version"], fg=_color, nl=False)
            click.secho(" (", nl=False)
            click.secho(finding["versions"], fg=_color, nl=False)
            click.secho(") via ", nl=False)
            click.secho(finding["source"], fg="cyan", nl=False)
            click.secho(" as ", nl=False)
            click.secho(finding["identifier"], fg="yellow", nl=False)
            click.secho(" ignored until ", nl=False)
            click.secho(finding["ignored"]["expires"], fg="cyan", nl=False)
            click.secho(".")

            click.secho(finding["ignored"]["reason"], fg="cyan")
            click.secho("-- ")
            continue

        click.secho("")
        click.secho(finding["name"], fg="white", nl=False)
        click.secho("==", nl=False)
        click.secho(finding["version"], fg=_color, nl=False)
        click.secho(" (", nl=False)
        click.secho(finding["versions"], fg=_color, nl=False)
        click.secho(") via ", nl=False)
        click.secho(finding["source"], fg="cyan", nl=False)
        click.secho(" as ", nl=False)
        click.secho(finding["identifier"], fg="yellow", nl=False)
        click.secho("")

        click.secho("")
        click.secho(textwrap.fill(finding["summary"], 79), fg="white")
        click.secho(finding["url"], fg="green")
        click.secho("")
        for reference in finding["references"]:
            click.secho(reference, fg="white")
        click.secho("-- ")

    # Always print the summary to stderr.
    if len(ignored_findings):
        click.secho(
            f"Ignored {len(ignored_findings)} finding(s)!", fg="yellow", err=True
        )
    if len(vulnerable_packages) > 0:
        click.secho(
            f"Found {len(vulnerable_packages)} vulnerable package(s)!",
            fg="red",
            blink=True,
            err=True,
        )
    else:
        click.secho(f"No vulnerable packages found!", fg="green", err=True)

    return vulnerable_packages, ignored_findings


def audit(
    configuration: Configuration,
    packages: PackageList,
    ignore: SkjoldIgnore,
) -> List[Dict[str, Any]]:
    """..."""

    findings = []
    for name in configuration.sources:
        source = _sources[name](
            cache_dir=configuration.cache_dir, cache_expires=configuration.cache_expires
        )

        for package_name, package_version in packages:
            if source.has_security_advisory_for(package_name):
                is_vulnerable, advisories = source.is_vulnerable_package(
                    package_name, package_version
                )

                if is_vulnerable:
                    for advisory in advisories:
                        # Check if the advisories identifier is part of the ignore list.
                        is_ignored, entry = ignore.should_ignore(
                            advisory.identifier, advisory.package_name
                        )
                        findings.append(
                            {
                                "identifier": advisory.identifier,
                                "severity": advisory.severity,
                                "name": package_name,
                                "version": package_version,
                                "versions": advisory.vulnerable_versions,
                                "source": source.name,
                                "summary": advisory.summary,
                                "references": advisory.references,
                                "url": advisory.url,
                                "ignored": {
                                    "ignored": is_ignored,
                                    "expires": entry.get("expires"),
                                    "reason": entry.get("reason"),
                                },
                            }
                        )

    return findings



import os
import tarfile
import urllib.request
from collections import defaultdict
from typing import List, Tuple

import yaml

from skjold.models import SecurityAdvisory, SecurityAdvisorySource, SkjoldException
from skjold.sources.osv import OSVSecurityAdvisory
from skjold.tasks import register_source


class PyPAAdvisoryDB(SecurityAdvisorySource):

    _url = "https://api.github.com/repos/pypa/advisory-db/tarball"
    _name = "pypa"

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return os.path.join(self._cache_dir, "pypa.cache")

    def populate_from_cache(self) -> None:
        self._advisories = defaultdict(list)
        with tarfile.TarFile.open(self.path, mode="r:gz") as archive:
            pypi_advisories = filter(
                lambda obj: "/vulns/" in obj.name and obj.name.endswith(".yaml"),
                archive.getmembers(),
            )

            for obj in list(pypi_advisories):
                obj_fh = archive.extractfile(obj.name)
                if obj_fh:
                    doc = yaml.load(obj_fh, Loader=yaml.SafeLoader)
                    advisory = OSVSecurityAdvisory.using(doc)
                    self._advisories[advisory.package_name.lower()].append(advisory)
                else:  # pragma: no cover
                    raise SkjoldException(
                        f"Unable to extract '{obj.name}' from source archive."
                    )

    @property
    def total_count(self) -> int:
        return len(self._advisories.keys())

    def update(self) -> None:
        request = urllib.request.Request(
            url=self._url, headers={"User-Agent": "Mozilla/5.0"}
        )
        with urllib.request.urlopen(request) as response:
            with open(self.path, "wb") as fh:
                fh.write(response.read())

    def has_security_advisory_for(self, package_name: str) -> bool:
        return package_name.strip().lower() in self.advisories.keys()

    def is_vulnerable_package(
        self, package_name: str, package_version: str
    ) -> Tuple[bool, List[SecurityAdvisory]]:
        if not self.has_security_advisory_for(package_name):
            return False, []

        advisories = []
        for candidate in self.advisories[package_name.strip().lower()]:
            if candidate.is_affected(package_version):
                advisories.append(candidate)

        return len(advisories) > 0, advisories


register_source("pypa", PyPAAdvisoryDB)



import json
import os
import urllib.request
from typing import List, Tuple, Callable, Dict, Any, Sequence
from typing import Optional, MutableMapping

import semver

from skjold.models import SecurityAdvisory, SecurityAdvisorySource, SecurityAdvisoryList
from skjold.tasks import register_source


def _is_supported_range_type(affected_range: Dict) -> bool:
    """Return True if the given `affected.ranges` item is supported."""
    return "type" in affected_range and affected_range["type"] in {
        "ECOSYSTEM",
        "SEMVER",
    }


def _osv_dev_api_request(
    package_name: str, package_version: str, ecosystem: str = "PyPI"
) -> Any:
    """Return list of vulnerabilities for a given `package_name` and `package_version` via OSV.dev API."""

    _api_token = os.environ.get("SKJOLD_OSV_API_TOKEN", None)

    payload = json.dumps(
        {
            "version": package_version,
            "package": {"name": package_name, "ecosystem": ecosystem},
        }
    ).encode("utf-8")
    request_ = urllib.request.Request(
        url="https://api.osv.dev/v1/query",
        data=payload,
        headers={
            "Accept": "application/json",
            # "Authorization": f"Bearer {_api_token}",
            "Content-Type": "application/json; charset=utf-8",
        },
    )
    with urllib.request.urlopen(request_) as response:
        _data = json.loads(response.read())

    return _data.get("vulns", [])


class OSVSecurityAdvisory(SecurityAdvisory):
    _json: dict

    @classmethod
    def using(cls, json_: dict) -> "OSVSecurityAdvisory":
        obj = cls()
        obj._json = json_
        return obj

    @property
    def identifier(self) -> str:
        return str(self._json["id"])

    @property
    def source(self) -> str:
        return "osv"

    @property
    def severity(self) -> str:
        return "UNKNOWN"

    @property
    def url(self) -> str:
        return str(self.references[0])

    @property
    def references(self) -> List[str]:
        return [str(reference["url"]) for reference in self._json["references"]]

    @property
    def package_name(self) -> str:
        return str(self._json["package"]["name"]).strip()

    @property
    def summary(self) -> str:
        return f"{self._json['details']}"

    @property
    def vulnerable_version_range(self) -> List[semver.VersionConstraint]:
        # Try using ranges first to avoid clutter.
        affected_ecosystem_ranges = list(
            filter(
                _is_supported_range_type,
                self._json.get("affects", {}).get("ranges", []),
            )
        )
        if len(affected_ecosystem_ranges):
            ranges = []
            for affected_range in affected_ecosystem_ranges:
                _constraints = []
                if "introduced" in affected_range:
                    v = affected_range.get("introduced")
                    _constraints.append(f">={v}")

                if "fixed" in affected_range:
                    v = affected_range.get("fixed")
                    _constraints.append(f"<{v}")

                ranges.append(semver.parse_constraint(",".join(_constraints)))
            return ranges

        # Try using versions (default).
        affected_versions = self._json.get("affects", {}).get("versions", [])
        if len(affected_versions):
            return [semver.parse_constraint(x) for x in affected_versions]

        return [semver.parse_constraint(">=0.0.0")]

    @property
    def vulnerable_versions(self) -> str:
        return "||".join([str(x) for x in self.vulnerable_version_range])

    def is_affected(self, version: str) -> bool:
        version_ = semver.Version.parse(version)
        allows_: Callable[[semver.VersionConstraint], bool] = (
            lambda x: True if x.allows(version_) else False
        )
        # affected_versions = map(lambda x: x.allows(version), self.vulnerable_version_range)
        affected_versions = map(allows_, self.vulnerable_version_range)
        return any(affected_versions)


class OSV(SecurityAdvisorySource):

    _name = "osv"

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> Optional[str]:
        return None

    def populate_from_cache(self) -> None:
        pass

    @property
    def total_count(self) -> int:
        return 0

    def update(self) -> None:
        pass

    def has_security_advisory_for(self, package_name: str) -> bool:
        """Always return `True` since to ensure we always call the OSV API for every package."""
        return True

    def is_vulnerable_package(
        self, package_name: str, package_version: str
    ) -> Tuple[bool, Sequence[SecurityAdvisory]]:

        findings = _osv_dev_api_request(package_name.strip().lower(), package_version)
        if not len(findings):
            return False, []

        advisories = []
        for finding in findings:
            advisory = OSVSecurityAdvisory.using(finding)
            advisories.append(advisory)

        return True, advisories

    def get_security_advisories(self) -> MutableMapping[str, SecurityAdvisoryList]:
        raise NotImplementedError


register_source("osv", OSV)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['skjold', 'skjold.sources']

package_data = \
{'': ['*']}

install_requires = \
['click>=8,<9',
 'poetry-semver>=0.1.0,<0.2.0',
 'pyyaml>=5.3,<6.0',
 'toml>=0.10.0']

entry_points = \
{'console_scripts': ['skjold = skjold.cli:cli']}

setup_kwargs = {
    'name': 'skjold',
    'version': '0.3.2',
    'description': 'Security audit Python project dependencies against security advisory databases.',
    'long_description': '![](https://img.shields.io/pypi/v/skjold?color=black&label=PyPI&style=flat-square)\n![](https://img.shields.io/github/workflow/status/twu/skjold/Python%20Package/master?color=black&label=Tests&style=flat-square)\n![](https://img.shields.io/pypi/status/skjold?color=black&style=flat-square)\n![](https://img.shields.io/pypi/pyversions/skjold?color=black&logo=python&logoColor=white&style=flat-square)\n![](https://img.shields.io/pypi/l/skjold?color=black&label=License&style=flat-square)\n![](https://img.shields.io/pypi/dm/skjold?color=black&label=Downloads&style=flat-square)\n[![](https://api.codeclimate.com/v1/badges/9f756df1ff145e6004a7/maintainability)](https://codeclimate.com/github/twu/skjold/maintainability)\n[![Test Coverage](https://api.codeclimate.com/v1/badges/9f756df1ff145e6004a7/test_coverage)](https://codeclimate.com/github/twu/skjold/test_coverage)\n\n```\n        .         .    .      Skjold /skjɔl/\n    ,-. | , . ,-. |  ,-|\n    `-. |<  | | | |  | |      Security audit python project dependencies\n    `-\' \' ` | `-\' `\' `-´      against several security advisory databases.\n           `\'\n```\n\n## Introduction\nIt currently supports fetching advisories from the following sources:\n\n| Source | Name | Notes |\n| ------:|:----:|:------|\n| [GitHub Advisory Database](https://github.com/advisories) | `github` | |\n| [PyUP.io safety-db](https://github.com/pyupio/safety-db) | `pyup` | |\n| [GitLab gemnasium-db](https://gitlab.com/gitlab-org/security-products/gemnasium-db) | `gemnasium` | |\n| [PYPA Advisory Database](https://github.com/pypa/advisory-db) | `pypa` | **Experimental!** Only supports `ECOSYSTEM` and `SEMVER`! |\n| [OSV.dev Database](https://osv.dev) | `osv` | **Experimental!** Only supports `ECOSYSTEM` and `SEMVER`!<br/> Sends package information to [OSV.dev](https://osv.dev) API. |\n\n\nNo source is enabled by default! Individual sources can be enabled by setting `sources` list (see [Configuration](#configuration)). There is (currently) no de-duplication meaning that using all of them could result in _a lot_ of duplicates.\n\n## Motivation\nSkjold was initially created for myself to replace `safety`. ~Which appears to no longer receive monthly updates (see [pyupio/safety-db #2282](https://github.com/pyupio/safety-db/issues/2282))~. I wanted something I can run locally and use for my local or private projects/scripts.\n\nI currently also use it during CI builds and before deploying/publishing containers or packages.\n\n## Installation\n`skjold` can be installed from either [PyPI](https://pypi.org/project/skjold/) or directly from [Github](https://github.com/twu/skjold) using `pip`:\n\n```sh\npip install skjold                                        # Install from PyPI\npip install git+https://github.com/twu/skjold.git@vX.X.X  # Install from Github\n```\n\nThis should provide a script named `skjold` that can then be invoked. See [Usage](#usage).\n\n## Usage\n```sh\n$ pip list --format=freeze | skjold -v audit --sources gemnasium -\n```\n\nWhen running `audit` one can either provide a path to a _frozen_ `requirements.txt`, a `poetry.lock` or a `Pipfile.lock` file. Alternatively, dependencies can also be passed in via `stdin`  (formatted as `package==version`).\n\n`skjold` will maintain a local cache (under `cache_dir`) that will expire automatically after `cache_expires` has passed. The `cache_dir` and `cache_expires` can be adjusted by setting them in  `tools.skjold` section of the projects `pyproject.toml` (see [Configuration](#configuration) for more details). The `cache_dir`will be created automatically, and by default unless otherwise specified will be located under `$HOME/.skjold/cache`.\n\nFor further options please read `skjold --help` and/or `skjold audit --help`.\n\n### Examples\n\nAll examples involving `github` assume that `SKJOLD_GITHUB_API_TOKEN` is already set (see [Github](#github)).\n\n```sh\n# Using pip list. Checking against GitHub only.\n$ pip list --format=freeze | skjold audit -s github -\n\n# Be verbose. Read directly from supported formats.\n$ skjold -v audit requirements.txt\n$ skjold -v audit poetry.lock\n$ skjold -v audit Pipenv.lock\n\n# Using poetry.\n$ poetry export -f requirements.txt | skjold audit -s github -s gemnasium -s pyup -\n\n# Using poetry, format output as json and pass it on to jq for additional filtering.\n$ poetry export -f requirements.txt | skjold audit -o json -s github - | jq \'.[0]\'\n\n# Using Pipenv, checking against Github\n$ pipenv run pip list --format=freeze | skjold audit -s github -\n\n# Checking a single package via stdin against Github and format findings as json.\n$ echo "urllib3==1.23" | skjold audit -o json -r -s github -\n[\n  {\n    "severity": "HIGH",\n    "name": "urllib3",\n    "version": "1.23",\n    "versions": "<1.24.2",\n    "source": "github",\n    "summary": "High severity vulnerability that affects urllib3",\n    "references": [\n      "https://nvd.nist.gov/vuln/detail/CVE-2019-11324"\n    ],\n    "url": "https://github.com/advisories/GHSA-mh33-7rrq-662w"\n  }\n]\n\n# Checking a single package via stdin against Gemnasium and report findings (`-o cli`).\n$ echo "urllib3==1.23" | skjold audit -o cli -r -s gemnasium -\n\nurllib3==1.23 (<=1.24.2) via gemnasium\n\nCRLF injection. In the urllib3 library for Python, CRLF injection is possible\nif the attacker controls the request parameter.\nhttps://nvd.nist.gov/vuln/detail/CVE-2019-11236\n--\n\nurllib3==1.23 (<1.24.2) via gemnasium\n\nWeak Authentication Caused By Improper Certificate Validation. The urllib3\nlibrary for Python mishandles certain cases where the desired set of CA\ncertificates is different from the OS store of CA certificates, which results\nin SSL connections succeeding in situations where a verification failure is the\ncorrect outcome. This is related to use of the `ssl_context`, `ca_certs`, or\n`ca_certs_dir` argument.\nhttps://nvd.nist.gov/vuln/detail/CVE-2019-11324\n--\n\nurllib3==1.23 (<1.25.9) via gemnasium\n\nInjection Vulnerability. urllib3 allows CRLF injection if the attacker controls\nthe HTTP request method, as demonstrated by inserting `CR` and `LF` control\ncharacters in the first argument of `putrequest()`. NOTE: this is similar to\nCVE-2020-26116.\nhttps://nvd.nist.gov/vuln/detail/CVE-2020-26137\n--\n\n# Ignore PYSEC-2020-148 finding from PyPA source until a certain date with a specific reason.\n$ skjold ignore urllib3 PYSEC-2020-148 --reason "Very good reason." --expires "2021-01-01T00:00:00+00:00"\nIgnore urllib3 in PYSEC-2020-148 until 2021-01-01 00:00:00+00:00?\nVery good reason.\n--\nAdd to \'.skjoldignore\'? [y/N]: y\n\n# Ignore PYSEC-2020-148 finding from PyPA source for 7 days with "No immediate remediation." reason.\n$ skjold ignore urllib3 PYSEC-2020-148\nIgnore urllib3 in PYSEC-2020-148 until ...?\nNo immediate remediation.\n--\nAdd to \'.skjoldignore\'? [y/N]: y\n\n# Audit `poetry.lock` using a custom `.skjoldignore` file location via `ENV`...\n$ SKJOLD_IGNORE_FILE=<path-to-file> skjold audit -s pyup poetry.lock\n\n# ... or using -i/--ignore-file\n$ skjold audit -s pyup -i <path-to-file> poetry.lock\n```\n\n### Configuration\n\n`skjold` can read its configuration from the `tools.skjold` section of a projects  `pyproject.toml`. Arguments specified via the command-line should take precedence over any configured or default value.\n\n```toml\n[tool.skjold]\nsources = ["github", "pyup", "gemnasium"]  # Sources to check against.\nreport_only = true                         # Report only, always exit with zero.\nreport_format = \'json\'                     # Output findings as `json`. Default is \'cli\'.\ncache_dir = \'.skjold_cache\'                # Cache location (default: `~/.skjold/cache`).\ncache_expires = 86400                      # Cache max. age.\nignore_file = \'.skjoldignore\'              # Ignorefile location (default `.skjoldignore`).\nverbose = true                             # Be verbose.\n```\n\nTo take a look at the current configuration / defaults run:\n```shell\n$ skjold config\nsources: [\'pyup\', \'github\', \'gemnasium\']\nreport_only: True\nreport_format: json\nverbose: False\ncache_dir: .skjold_cache\ncache_expires: 86400\nignore_file = \'.skjoldignore\'\n```\n\n#### Github\n\nFor the `github` source to work you\'ll need to provide a Github API Token via an `ENV` variable named `SKJOLD_GITHUB_API_TOKEN`. You can [create a new Github Access Token here](https://github.com/settings/tokens). You *do not* have to give it *any* permissions as it is only required to query the [GitHub GraphQL API v4](https://developer.github.com/v4/) API.\n\n### Version Control Integration\nTo use `skjold` with the excellent [pre-commit](https://pre-commit.com/) framework add the following to the projects `.pre-commit-config.yaml` after [installation](https://pre-commit.com/#install).\n\n```yaml\nrepos:\n  - repo: https://github.com/twu/skjold\n    rev: vX.X.X\n    hooks:\n    - id: skjold\n      verbose: true  # Important if used with `report_only`, see below.\n```\n\nAfter running `pre-commit install` the hook should be good to go. To configure `skjold` in this scenario I recommend adding the entire configuration to the projects `pyproject.toml` instead of manipulating the hook `args`. See this projects [pyproject.toml](./pyproject.toml) for an example.\n\n> **Important!**: When using `skjold` as a `pre-commit`-hook it only gets triggered if you want to commit changed dependency files (e.g. `Pipenv.lock`, `poetry.lock`, `requirements.txt`,...).\n> It will not continuously check your dependencies on _every_ commit!\n\nYou could run `pre-commit run skjold --all-files` manually in your workflow/scripts or run `skjold` manually.\nIf you have a better solution please let me know!\n\n> **Important!**: If you use `report_only` in any way make sure that you add `verbose: true` to your hook configuration\notherwise `pre-commit` won\'t show you any output since the hook is always returning with a zero exit code due\nto `report_only` being set!\n\n## Contributing\nPull requests are welcome! For major changes, please open an issue first to discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n',
    'author': 'Thomas Wurmitzer',
    'author_email': 't.wurmitzer+github@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/twu/skjold',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import datetime
import os
from typing import Dict, Tuple

import yaml


class SkjoldIgnore:
    _doc: Dict
    _path: str

    EXPIRES_FMT = "%Y-%m-%dT%H:%M:%S%z"
    DEFAULT_EXPIRES = datetime.datetime.now() + datetime.timedelta(days=7)
    DEFAULT_REASON = "No immediate remediation."

    def __init__(self, path: str):
        self._path = path
        self._doc = {"version": "1.0", "ignore": {}}

    @classmethod
    def using(cls, path: str) -> "SkjoldIgnore":
        obj = SkjoldIgnore(path)
        if not os.path.exists(path):
            return obj

        with open(path) as fh:
            obj._doc = yaml.safe_load(fh)
        return obj

    @property
    def version(self) -> str:
        return str(self._doc["version"])

    @property
    def entries(self) -> Dict:
        entries = self._doc["ignore"]
        return dict(entries)

    def add(
        self,
        identifier: str,
        package_name: str,
        reason: str,
        expires: datetime.datetime = datetime.datetime.now()
        + datetime.timedelta(days=14),
    ) -> bool:

        expires = expires.replace(tzinfo=datetime.timezone.utc)

        if identifier not in self.entries:
            self._doc["ignore"][identifier] = []

        self._doc["ignore"][identifier].append(
            {
                "package": package_name,
                "reason": reason,
                "expires": expires.strftime(SkjoldIgnore.EXPIRES_FMT),
            }
        )
        return True

    def save(self) -> None:
        with open(self._path, "w") as fh:
            yaml.safe_dump(self._doc, fh)

    def should_ignore(self, identifier: str, package_name: str) -> Tuple[bool, Dict]:
        """Returns True a given identifier has an entry in the blacklist."""
        if identifier not in self.entries:
            return False, {}

        for entry in self.entries.get(identifier, {}):
            if entry["package"] == package_name:
                dt = datetime.datetime.strptime(
                    f"{entry['expires']}", SkjoldIgnore.EXPIRES_FMT
                )
                return datetime.datetime.now(tz=dt.tzinfo) < dt, entry

        return False, {}



import os
import tarfile
import urllib.request
from collections import defaultdict
from typing import List, Tuple, Callable

import semver
import yaml

from skjold.cvss import parse_cvss
from skjold.models import SecurityAdvisory, SecurityAdvisorySource, SkjoldException
from skjold.tasks import register_source


class GemnasiumSecurityAdvisory(SecurityAdvisory):
    _json: dict

    @classmethod
    def using(cls, json_: dict) -> "GemnasiumSecurityAdvisory":
        obj = cls()
        obj._json = json_
        return obj

    @property
    def identifier(self) -> str:
        return str(self._json["identifier"])

    @property
    def source(self) -> str:
        return "gemnasium"

    @property
    def severity(self) -> str:

        for field in ["cvss_v3", "cvss_v2"]:
            vector = self._json.get(field, None)
            if vector:
                cvss = parse_cvss(vector)
                return cvss.severity

        return "UNKNOWN"

    @property
    def url(self) -> str:
        return str(self.references[0])

    @property
    def references(self) -> List[str]:
        return [str(url) for url in self._json["urls"]]

    @property
    def package_name(self) -> str:
        return str(self._json["package_slug"]).replace("pypi/", "").strip()

    @property
    def summary(self) -> str:
        return f"{self._json['title']}. {self._json['description']}"

    @property
    def vulnerable_version_range(self) -> List[semver.VersionConstraint]:
        affected_range = self._json["affected_range"]

        if not affected_range:
            return [semver.parse_constraint(">=0.0.0")]

        return [semver.parse_constraint(x) for x in affected_range.split("||")]

    @property
    def vulnerable_versions(self) -> str:
        return ",".join([str(x) for x in self.vulnerable_version_range])

    def is_affected(self, version: str) -> bool:
        version_ = semver.Version.parse(version)
        allows_: Callable[[semver.VersionConstraint], bool] = (
            lambda x: True if x.allows(version_) else False
        )
        # affected_versions = map(lambda x: x.allows(version), self.vulnerable_version_range)
        affected_versions = map(allows_, self.vulnerable_version_range)
        return any(affected_versions)


class Gemnasium(SecurityAdvisorySource):

    _url = "https://gitlab.com/gitlab-org/security-products/gemnasium-db/-/archive/master/gemnasium-db-master.tar.gz"
    _name = "gemnasium"

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return os.path.join(self._cache_dir, "gemnasium.cache")

    def populate_from_cache(self) -> None:
        self._advisories = defaultdict(list)
        with tarfile.TarFile.open(self.path, mode="r:gz") as archive:
            pypi_advisories = filter(
                lambda obj: "/pypi/" in obj.name and obj.name.endswith(".yml"),
                archive.getmembers(),
            )

            for obj in list(pypi_advisories):
                obj_fh = archive.extractfile(obj.name)
                if obj_fh:
                    doc = yaml.load(obj_fh, Loader=yaml.SafeLoader)
                    advisory = GemnasiumSecurityAdvisory.using(doc)
                    self._advisories[advisory.package_name.lower()].append(advisory)
                else:  # pragma: no cover
                    raise SkjoldException(
                        f"Unable to extract '{obj.name}' from source archive."
                    )

    @property
    def total_count(self) -> int:
        return len(self._advisories.keys())

    def update(self) -> None:
        request = urllib.request.Request(
            url=self._url, headers={"User-Agent": "Mozilla/5.0"}
        )
        with urllib.request.urlopen(request) as response:
            with open(self.path, "wb") as fh:
                fh.write(response.read())

    def has_security_advisory_for(self, package_name: str) -> bool:
        return package_name.strip().lower() in self.advisories.keys()

    def is_vulnerable_package(
        self, package_name: str, package_version: str
    ) -> Tuple[bool, List[SecurityAdvisory]]:
        if not self.has_security_advisory_for(package_name):
            return False, []

        advisories = []
        for candidate in self.advisories[package_name.strip().lower()]:
            if candidate.is_affected(package_version):
                advisories.append(candidate)

        return len(advisories) > 0, advisories


register_source("gemnasium", Gemnasium)



#!/usr/bin/env python
__version__ = "0.3.2"

import datetime
import os
import sys
from typing import List, TextIO

import click
import skjold.sources

from skjold.formats import extract_package_list_from, Format
from skjold.ignore import SkjoldIgnore
from skjold.tasks import (
    Configuration,
    audit,
    print_configuration,
    report,
    get_registered_sources,
    default_from_context,
    get_configuration_from_toml,
)

configuration = click.make_pass_decorator(Configuration, ensure=True)


@click.group()  # pragma: no cover
@click.option(
    "configuration_file",
    "-c",
    "--configuration-file",
    envvar="SKJOLD_RC",
    type=click.Path(exists=False, dir_okay=False, resolve_path=True),
    default="./pyproject.toml",
    show_default=True,
    required=False,
)
@click.option(
    "verbose",
    "-v",
    "--verbose",
    envvar="SKJOLD_VERBOSE",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.version_option(version=__version__, prog_name="skjold")
@configuration
def cli(
    config: Configuration,
    configuration_file: click.Path,
    verbose: bool,
) -> None:
    """Check a given Python dependency file against a set of advisory databases."""
    config.verbose = verbose

    file_ = str(configuration_file)
    skip_configuration = not os.environ.get("SKJOLD_SKIP_RC", None) is None

    if os.path.exists(file_) and not skip_configuration:
        settings = get_configuration_from_toml(file_)
        config.use(config=settings)
    else:
        click.secho("Warning: No 'pyproject.toml' found!", err=True, fg="yellow")

    if config.verbose:
        print_configuration(config)
        click.secho(f"Using {config.cache_dir} as cache location", err=True)
        click.secho(f"Using {config.ignore_file} as ignore file", err=True)

    # Cache Directory
    # Check for cache directory and create it if necessary.
    if not os.path.isdir(config.cache_dir):
        os.makedirs(config.cache_dir, exist_ok=True)
        if config.verbose:
            click.secho(
                f"Cache '{config.cache_dir}' does not exist! Creating it.",
                err=True,
            )

    if not os.path.isdir(config.cache_dir):
        raise click.ClickException(
            f"Unable to create cache directory '{config.cache_dir}'!"
        )


@cli.command("config")  # pragma: no cover
@configuration
def config_(config: Configuration) -> None:
    """Print the current configuration and exit."""
    print_configuration(config)


@cli.command("audit")  # pragma: no cover
@click.option(
    "report_only",
    "-r",
    "--report-only",
    is_flag=True,
    cls=default_from_context("report_only", Configuration),
    help="Only report findings, always exit with zero.",
    show_default=True,
)
@click.option(
    "report_format",
    "-o",
    "--report-format",
    type=click.Choice(["json", "cli"], case_sensitive=True),
    cls=default_from_context("report_format", Configuration),
    help="Output format",
    show_default=True,
)
@click.option(
    "file_format",
    "-f",
    "--file-format",
    type=click.Choice(Format.SUPPORTED_FORMATS, case_sensitive=True),
    default=Format.REQUIREMENTS,
    help="Input format",
    show_default=True,
)
@click.option(
    "ignore_file",
    "-i",
    "--ignore-file",
    type=str,
    cls=default_from_context("ignore_file", Configuration),
    help="Ignore file location.",
    show_default=True,
)
@click.option(
    "sources",
    "-s",
    "--sources",
    type=click.Choice(get_registered_sources(), case_sensitive=True),
    cls=default_from_context("sources", Configuration),
    help="Identifier of a registered advisory source.",
    show_default=False,
    multiple=True,
)
@click.argument("file", type=click.File(), default="./requirements.txt", required=False)
@configuration
def audit_(
    config: Configuration,
    report_only: bool,
    report_format: str,
    file_format: str,
    ignore_file: str,
    sources: List[str],
    file: TextIO,
) -> None:
    """
    Checks a given dependency file against advisory databases.

    \b
    FILE is the path to the dependency file to audit.
    """
    config.report_only = report_only
    config.report_format = report_format
    config.ignore_file = ignore_file

    # Only override sources if at least once --source is passed.
    if len(sources) > 0:
        config.sources = list(set(sources))

    if len(config.sources) == 0:
        raise click.ClickException(
            "Please specify or configure at least one advisory source."
        )

    packages = extract_package_list_from(config, file, file_format)

    if config.verbose:
        click.secho("Checking ", nl=False, err=True)
        click.secho(f"{len(packages)}", fg="green", nl=False, err=True)
        click.secho(" package(s).", err=True)

        click.secho("Using ", nl=False, err=True)
        click.secho(f"{config.sources}", fg="green", nl=False, err=True)
        click.secho(" as source(s).", err=True)

    ignore = SkjoldIgnore.using(config.ignore_file)

    findings = audit(config, packages, ignore=ignore)

    vulnerable_packages, _ = report(config, findings)

    # By default we want to exit with a non-zero exit-code when we encounter
    # any findings.
    if not config.report_only and len(vulnerable_packages) > 0:
        sys.exit(1)


@cli.command("ignore")  # pragma: no cover
@click.option(
    "reason",
    "-r",
    "--reason",
    type=str,
    default=SkjoldIgnore.DEFAULT_REASON,
    help="Reason for the finding to be ignored.",
    show_default=True,
)
@click.option(
    "expires",
    "-e",
    "--expires",
    type=click.DateTime(formats=[SkjoldIgnore.EXPIRES_FMT]),
    help="Ignore finding until after this date.",
    default=SkjoldIgnore.DEFAULT_EXPIRES,
    show_default=True,
)
@click.argument("package", type=str, required=True)
@click.argument("identifier", type=str, required=True)
@configuration
def ignore_(
    config: Configuration,
    reason: str,
    expires: datetime.datetime,
    package: str,
    identifier: str,
) -> None:
    """
    Adds a finding with a given source identifier and package name to a `.skjoldignore` file.

    \b
    IDENTIFIER The vulnerability identifier to ignore e.g. CVE-2021-02231
    PACKAGE The name of the package for which this identifier should be applied to.
    """
    click.secho("Ignore ", nl=False)
    click.secho(package, fg="red", nl=False)
    click.secho(" in ", nl=False)
    click.secho(identifier, fg="red", nl=False)
    click.secho(" until ", nl=False)
    click.secho(expires, fg="red", nl=False)
    click.secho("?")

    click.secho(reason, fg="yellow")
    click.secho("-- ")

    if click.confirm(f"Add to '{config.ignore_file}'?"):
        ignore = SkjoldIgnore.using(config.ignore_file)
        ignore.add(identifier, package, reason=reason, expires=expires)
        ignore.save()


if __name__ == "__main__":
    cli()
